"""
Amitoz AZAD 2022-04-19 06:59
"""

import os
import argparse

import matplotlib.pyplot as plt

from pyinpaint import Inpaint


def modify(path):
    file_name = os.path.splitext(path)[0]
    file_extension = os.path.splitext(path)[1]
    path = file_name +"_inpainted" + file_extension
    return path


def main():
    """
    Calls the inpaint module, and saves the inpainted image.
    """
    argparser = argparse.ArgumentParser()
    argparser.add_argument("--org_img", type=str, help="The path to the original image.")
    argparser.add_argument("--mask", type=str, help="The path to the original image.")
    argparser.add_argument("--ps", type=int, default=7, help="Patch size.")
    argparser.add_argument("--k_boundary", type=int, default=4, help="To determine the boundary pixels. Ideally should be 4 or 8.")
    argparser.add_argument("--k_search", type=int, default=1000, help="Determines the search range, normally 500 or 1000")
    argparser.add_argument("--k_patch", type=int, default=5, help="knn value for the non-local graph, ideal values are 3,5,7")

    args = argparser.parse_args()

    inpaint = Inpaint(args.org_img, args.mask, args.ps)
    inpainted_img = inpaint(args.k_boundary, args.k_search, args.k_patch)

    path_to_output = modify(args.org_img) 
    plt.imsave(path_to_output, inpainted_img)


if __name__ == "__main__":
    main()



"""
Amitoz AZAD 2022-04-18 11:29
"""


# load libs
import numpy as np 
from scipy import spatial
from tqdm import tqdm
import matplotlib.image as mpimg

from pyinpaint.utils import *


class Inpainting:
  """
  Inpaints an image using Dirichlet interpolation.
  Happens to be nlm for p=2 in p laplacian.
  """
  def __init__(self, org_img, mask, ps):
    """
    org_img: path to the original image (rgb or gray)
    mask: path to the mask (binary image, 0,1)
    ps: patch size (used for creating a dynamic non-local graph)
    """
    self.org_img = org_img
    self.mask = mask 
    self.ps = ps

  def __call__(self, k_boundary=4, k_search=1000, k_patch=5):
    """
    k_boundary: used for finding the boundary pixels
    k_search: used for search in the neighborhood of the boundary pixels
    k_patch: used for creating a dynamic non-local graph
    """
    inpainted_img =  self.forward(k_boundary, k_search, k_patch)
    return inpainted_img

  def preprocess(self):
    img = mpimg.imread(self.org_img)
    mask = mpimg.imread(self.mask)
    img = (img - np.min(img))/ (np.max(img) - np.min(img)).astype("float32") + 0.01
    img = (img.T * mask.T).T
    self._shape = img.shape
    position = pmat(self._shape)
    texture = fmat(img)
    self._position = (position - np.min(position) )/ (np.max(position) - np.min(position))
    self._texture = (texture - np.min(texture) )/ (np.max(texture) - np.min(texture))
    self._patches = create_patches(img, (self.ps, self.ps))

  def postprocess(self, fmat):
      return to_img(fmat, self._shape)

  def forward(self, k_boundary, k_search, k_patch):
    self.preprocess()

    kdt = spatial.cKDTree(self._position)
    dA = np.where(self._texture.any(axis=1))[0]
    A = np.where(~self._texture.any(axis=1))[0]

    pbar = tqdm(desc=f"# of pixels to be inpainted are {A.size}", total = A.size, 
            bar_format = '{l_bar}{bar}|{n_fmt}/{total_fmt}')
    while A.size >=1 :
      dmA = np.array([]).astype("int")
      for i in A : 
        _, indices = kdt.query(self._position[i], k_boundary)
        if (~np.isin(indices, A)).any():
          dmA = np.append(dmA,i)
          mask = (~(self._patches[i].flatten() == 0)).astype("int")
          _, indices = kdt.query(self._position[i], k_search)
          part_of_dA = indices[~np.isin(indices,A)]
          new_patches = mask.flatten() * self._patches[part_of_dA]
          kdt_ = spatial.cKDTree(new_patches)
          _, indices= kdt_.query(self._patches[i].flatten(),k_patch)
          ids = part_of_dA[indices]
          self._texture[i] = self._texture[ids].mean(axis=0)
      self._patches = create_patches(to_img(self._texture,(self._shape)),(self.ps,self.ps))
      dA = np.concatenate((dA,dmA), axis=0)
      A = A[~np.isin(A,dmA)]
      #pbar.set_description(desc=f"# of pixels to be inpainted are {A.size}")
      pbar.update(dmA.size)
    pbar.close()
    return self.postprocess(self._texture)


def check_args(cls):
  def correct_args(org_img, mask, ps=7):
    if (isinstance(org_img,str) and isinstance(mask,str) and isinstance(ps,int)):
      return cls(org_img, mask, ps)
    else:
      raise Exception(f"arg[0]:str, arg[1]:str, arg[2]:int")
  return correct_args


Inpaint = check_args(Inpainting)



"""
Amitoz AZAD 2022-04-18 10:14
"""

from setuptools import setup, find_packages

install_requires = [
    'numpy',
    'tqdm',
    'scipy',
    'scikit-image',
    'matplotlib',
]

setup(
    name="pyinpaint",
    version="0.0.3",
    description="A lightweight image inpainting tool",
    url="https://github.com/aGIToz/PyInpaint",
    python_requires='>=3.6',
    install_requires=install_requires,
    packages=find_packages(),
    author="Amitoz AZAD",
    license="MIT",
    include_package_data=True,
    entry_points={
        "console_scripts": [
        "pyinpaint=pyinpaint.__main__:main",
        ]
        },
    )



"""
Amitoz AZAD 2022-04-18 11:30
Basic utilities to preprocess and postprocess.
"""


# libs
import numpy as np
import skimage


def fmat(img):
    """
    Converts to a feature matrix (N by c)
    """
    try:
        l, w, c = img.shape
        fmat = np.ones((l*w,c)) 
        for i in range(c): fmat[:,i] = np.reshape(img[:,:,i],(l*w,))
    except ValueError:
        try:
            l, w = img.shape
            fmat = np.ones((l*w,1)) 
            fmat = np.reshape(img,(l*w,1))
        except ValueError:
            print("Image should be at least a 2D array.")
    return fmat



def pmat(shape):
    """
    Returns the position feature matrix.
    """
    x = np.arange(0,shape[1],1)
    y = np.arange(shape[0],0,-1)
    meshx, meshy = np.meshgrid(x,y)
    x = np.reshape(meshx,(shape[0]*shape[1],1)) 
    y = np.reshape(meshy,(shape[0]*shape[1],1)) 
    pmat = np.concatenate((x,y),axis=1) / max(shape[0],shape[1])
    return pmat



def to_img(fmat, shape):
    """
    Converts from fmat to img.
    """
    img = np.zeros(shape)
    try:
        l, w, c = img.shape
        for i in range(c): img[:,:,i] = np.reshape(fmat[0:,i],(l,w))
    except ValueError:
        try:
            l, w = img.shape
            img = np.reshape(fmat,(l,w))
        except ValueError:
            print("Image should be at least a 2D array.")
    return img


# code taken from pygsp https://github.com/epfl-lts2/pygsp
# thanks to Michaël Defferrard 
def create_patches(img, patch_shape=(3,3)):
    try:
        h, w, d = img.shape
    except ValueError:
        try:
            h, w = img.shape
            d = 0
        except ValueError:
            print("Image should be at least a 2D array.")

    try:
        r, c = patch_shape
    except ValueError:
        r = patch_shape[0]
        c = r

    pad_width = [(int((r - 0.5) / 2.), int((r + 0.5) / 2.)),
                 (int((c - 0.5) / 2.), int((c + 0.5) / 2.))]

    if d == 0:
        window_shape = (r, c)
        d = 1  # For the reshape in the return call
    else:
        pad_width += [(0, 0)]
        window_shape = (r, c, d)

    # Pad the image.
    img = np.pad(img, pad_width=pad_width, mode='symmetric')

    # Extract patches as node features.
    # Alternative: sklearn.feature_extraction.image.extract_patches_2d.
    #              sklearn has much less dependencies than skimage.
    try:
        import skimage
    except Exception:
        raise ImportError('Cannot import skimage, which is needed to '
                          'extract patches. Try to install it with '
                          'pip (or conda) install scikit-image.')
    patches = skimage.util.view_as_windows(img, window_shape=window_shape)
    patches = patches.reshape((h * w, r * c * d))
    return patches
